/* Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pfft.h"
#include "ipfft.h"


/* wrappers for fftw init and cleanup */
void PX(init) (void){
  XM(init)();
}

void PX(cleanup) (void){
  XM(cleanup)();
}


/* functions to execute and destroy PX(plan) */

void PX(execute)(PX(plan) ths)
{
  if(ths==NULL){
    PX(fprintf)(MPI_COMM_WORLD, stderr, "!!! Error: Can not execute PFFT Plan == NULL !!!\n");
    return;
  }

  ths->timer.whole -= MPI_Wtime();
  if(ths->fft_last_dims != NULL){
    ths->timer.fft[3] -= MPI_Wtime();
    X(execute)(ths->fft_last_dims);
    ths->timer.fft[3] += MPI_Wtime();
  }
  
  for(int p=0; p<3; p++){
    ths->timer.fft[p] -= MPI_Wtime();
    PX(serfft_trunc_execute)(ths->local_fft[p]);
    ths->timer.fft[p] += MPI_Wtime();

    if(ths->global_remap[p] != NULL){
      ths->timer.remap[p] -= MPI_Wtime();
      X(execute)(ths->global_remap[p]);
      ths->timer.remap[p] += MPI_Wtime();
    }
  }
  
  ths->timer.iter++;
  ths->timer.whole += MPI_Wtime();
} /* end PX(execute_timed) */


void PX(destroy_gcplan) (PX(gcplan) ths){
  PX(destroy_gcplan_internal)(ths);
}


void PX(destroy_plan) (PX(plan) ths){
  int p;

  if(ths==NULL){
    PX(fprintf)(MPI_COMM_WORLD, stderr, "!!! Error: Can not finalize PFFT Plan == NULL !!!\n");
    return;
  }

  if(ths->fft_last_dims != NULL)
    X(destroy_plan)(ths->fft_last_dims);

  for(p=0; p<3; p++){
    if(ths->local_fft[p] != NULL)
      PX(serfft_trunc_destroy_plan)(ths->local_fft[p]);
    if(ths->global_remap[p] != NULL)
      X(destroy_plan)(ths->global_remap[p]);
  }
  
  PX(free)(ths->n);
  PX(free)(ths);
} /* end PX(destroy_plan) */


void PX(init_input_3d)(
    const INT *n, const INT *local_n, const INT *local_n_start, MPI_Comm comm,
    C *data
    )
{ /* initialize FFT input with random numbers */
  INT howmany = 1;
  
  PX(init_many_input_3d)(
      n, n, howmany, local_n, local_n_start, comm,
      data);
}

void PX(check_output_3d)(
    C *data, const INT *n,
    const INT *local_n, const INT *local_n_start,
    MPI_Comm comm, int iterations, int is_scaled
    )
{ /* check results after one forward and backward FFT, only works for intial data layout */
  INT howmany = 1;
  
  PX(check_many_output_3d)(
      data, n, n, howmany, local_n, local_n_start, comm, iterations, is_scaled);
}

INT PX(local_size_3d)(
    const INT *n,
    MPI_Comm comm_cart_2d, unsigned pfft_flags,
    INT *local_ni, INT *local_i_start,
    INT *local_no, INT *local_o_start
    )
{
  int rank = 3;
  
  return PX(local_size)(rank, n, comm_cart_2d, pfft_flags,
      local_ni, local_i_start, local_no, local_o_start);
}

INT PX(local_size)(
    int rank, const INT *n,
    MPI_Comm comm_cart_2d, unsigned pfft_flags,
    INT *local_ni, INT *local_i_start,
    INT *local_no, INT *local_o_start
    )
{
  INT howmany = 1; /* howmany>1 only for d>3 */

  return PX(local_size_many)(rank, n, n, n, howmany,
      PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS, comm_cart_2d, pfft_flags,
      local_ni, local_i_start, local_no, local_o_start);
}


INT PX(local_size_gcells_3d)(
    const INT *local_n, const INT *local_start,
    INT alloc_local, const INT *num_gcells_below, const INT *num_gcells_above,
    INT *local_ngc, INT *local_gc_start
    )
{
  int rank = 3;
  
  return PX(local_size_gcells)(rank, local_n, local_start,
      alloc_local, num_gcells_below, num_gcells_above,
      local_ngc, local_gc_start);
}

INT PX(local_size_gcells)(
    int rank, const INT *local_n, const INT *local_start,
    INT alloc_local, const INT *num_gcells_below, const INT *num_gcells_above,
    INT *local_ngc, INT *local_gc_start
    )
{
  INT howmany = 1;
  
  return PX(local_size_many_gcells)(rank, local_n, local_start,
      alloc_local, howmany, num_gcells_below, num_gcells_above,
      local_ngc, local_gc_start);
}


PX(plan) PX(plan_dft_3d)(
    const INT *n, C *data_in, C *data_out,
    MPI_Comm comm_cart_2d,
    int sign, unsigned pfft_flags, unsigned fftw_flags
    )
{
  int rank = 3;
  
  return PX(plan_dft)(rank, n, data_in, data_out, comm_cart_2d,
      sign, pfft_flags, fftw_flags);
}


PX(plan) PX(plan_dft)(
    int rank, const INT *n, C *data_in, C *data_out,
    MPI_Comm comm_cart_2d,
    int sign, unsigned pfft_flags, unsigned fftw_flags
    )
{ /* !!! with flag FFTW_MEASURE planning will overwrite users data !!! */
  INT howmany = 1;

  return PX(plan_many_dft)(
      rank, n, n, n, howmany, PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS,
      data_in, data_out, comm_cart_2d, sign, pfft_flags, fftw_flags);
}


PX(gcplan) PX(plan_gcells_3d)(
    const INT *n,
    const INT *num_gcells_below, const INT *num_gcells_above,
    C *data, MPI_Comm comm_cart_2d, unsigned pfft_flags
    )
{
  int rank = 3;
  
  return PX(plan_gcells)(rank, n, num_gcells_below, num_gcells_above,
      data, comm_cart_2d, pfft_flags);
}


PX(gcplan) PX(plan_gcells)(
    int rank, const INT *n,
    const INT *num_gcells_below, const INT *num_gcells_above,
    C *data, MPI_Comm comm_cart_2d, unsigned pfft_flags
    )
{
  INT howmany = 1;
  
  return PX(plan_many_gcells)(rank, n, howmany, PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS,
      num_gcells_below, num_gcells_above, data, comm_cart_2d, pfft_flags);
}




int PX(check_3d)(
    const INT *n, MPI_Comm comm_cart_2d, unsigned pfft_flags, int verbose
    )
{
  return PX(check_many_3d)(n, n, n, PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS, comm_cart_2d, pfft_flags, verbose);
}


void PX(exchange)(
    PX(gcplan) ths
    )
{
  ths->timer_exg.iter++;
  ths->timer_exg.whole -= MPI_Wtime();
  PX(exchange_gcells_internal)(ths);
  ths->timer_exg.whole += MPI_Wtime();
}


void PX(reduce)(
    PX(gcplan) ths
    )
{
  ths->timer_red.iter++;
  ths->timer_red.whole -= MPI_Wtime();
  PX(reduce_gcells_internal)(ths);
  ths->timer_red.whole += MPI_Wtime();
}



static void print_complex_array(
     const double *data, const INT *n, const INT *start, const char *name
     )
{
  INT k0, k1, k2, l;
  
  printf("%s(%td:%td,%td:%td,%td:%td):", name, start[0],start[0]+n[0]-1, start[1],
      start[1]+n[1]-1, start[2], start[2]+n[2]-1);
  for(k0 = 0, l = 0; k0 < n[0]; k0++){
    for(k1 = 0; k1 < n[1]; k1++){
      for(k2 = 0; k2 < n[2]; k2++, l++){
        if(k2%4 == 0)
          printf("\n%4td.", l);
        printf(" %.2e+%.2ei,", data[2*l], data[2*l+1]);
      }
    }
    printf("\n");
  }
}


void PX(apr_complex_3d)(
     const fftw_complex *data, const INT *local_n, const INT *local_n_start,
     const char *name, MPI_Comm comm
     )
{
  int num_procs, proc_rank;
  
  if( PX(prod_INT)(3, local_n) < 1)
    return;

  MPI_Comm_size(comm, &num_procs);
  MPI_Comm_rank(comm, &proc_rank);

  fflush(stdout);
  for(int t=0; t<num_procs; t++){
    if(proc_rank == t){
      printf("\nRank %d: ", proc_rank);
      print_complex_array((double*) data, local_n, local_n_start, name);
      fflush(stdout);
    }
    MPI_Barrier(comm);
  }
}


void PX(apr_complex_permuted_3d)(
     const fftw_complex *data, const INT *local_n, const INT *local_n_start,
     int perm0, int perm1, int perm2,
     const char *name, MPI_Comm comm
     )
{
  int perm[3];
  INT local_n_start_perm[3], local_n_perm[3];
  
  perm[0] = perm0; perm[1] = perm1; perm[2] = perm2;
  
  for(int t=0; t<3; t++){
    local_n_perm[t] = local_n[perm[t]];
    local_n_start_perm[t] = local_n_start[perm[t]];
  }
  
  PX(apr_complex_3d)(data, local_n_perm, local_n_start_perm, name, comm);
}

