/*
 * Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pfft.h"
#include "ipfft.h"

static serfft_trunc_plan serfft_trunc_mkplan(
    void);
static serfft_plan plan_dft_1d(
    const INT *local_n, INT howmany, C *data_in, C *data_out,
    int sign, unsigned transpose_flags, unsigned fftw_flags,
    PX(permutation) *perm);
static serfft_plan plan_serfft(
    INT *n_in, INT *is_in, INT *os_in, int dim_fft, INT howmany, C *data_in, C *data_out,
    int sign, unsigned fftw_flags);
static serfft_plan plan_serfft_p(
    INT *n, INT *is, INT *os, INT howmany,
    C *data_in0, C *data_out0, C *data_in1, C *data_out1,
    int sign, unsigned fftw_flags, double *time, unsigned *algorithm_flag);
static serfft_plan plan_serfft_ps(
    INT *n, INT *is0, INT *os0, INT *is1, INT *os1, INT howmany,
    C *data_in0, C *data_out0, C *data_in1, C *data_out1,
    int sign, unsigned fftw_flags, double *time, unsigned *algorithm_flag);
static void print_algorithm_flag(
    unsigned flag);
static double measure_time(
    X(plan) plan);
static serfft_plan serfft_mkplan(
    void);
static void serfft_destroy_plan(
    serfft_plan serfft);
static X(plan) plan_serfft_wrapper(
    INT *n, INT *is, INT *os, INT howmany, C *data_in, C *data_out, int sign, unsigned fftw_flags);
static X(plan) plan_remap_wrapper(
    INT *n, INT *is, INT *os, INT howmany, C *data_in, C *data_out, unsigned fftw_flags);
static void fill_dims(
    INT *n, INT *is, INT *os, INT howmany,
    X(iodim64) *dims);
static void embed(
    INT *n, INT *no,
    INT howmany, C *data);
static void truncate(
    INT *n, INT *no,
    INT howmany, C *data);


void PX(permute_fft_1d)(
    unsigned fft_flag, PX(permutation) *perm
    )
{
  PX(permute_vector)(fft_flag, perm->n);
  PX(permute_vector)(fft_flag, perm->np);
}


void PX(mark_performed_fft)(
    const PX(permutation) *perm,
    PX(truncation) *trc
    )
{
  int dim_fft = PX(get_local_dim)(perm);
  trc->is_transformed[perm->n[dim_fft]]++;
}


INT PX(local_size_fft_1d)(
    const INT *local_n, INT howmany, unsigned fft_flag,
    PX(truncation) *trc, PX(permutation) *perm
    )
{
  INT alloc_local = howmany;

  for(int t=0; t<3; t++)
    alloc_local *= local_n[t];

  PX(permute_fft_1d)(fft_flag, perm);
  PX(mark_performed_fft)(perm, trc);

  return alloc_local;
}



void PX(serfft_trunc_execute)(
    serfft_trunc_plan plan)
{
  embed(plan->local_ni_embed, plan->local_n_embed, plan->howmany, plan->data_in);
  PX(serfft_execute)(plan->serfft);
  truncate(plan->local_n_trunc, plan->local_no_trunc, plan->howmany, plan->data_out);
}


static void embed(
    INT *n, INT *no,
    INT howmany, C *data
    )
{
  INT zeros_size[3], zeros_start[3];

  for(int t=0; t<3; t++)
    zeros_start[t] = (n[t]+1)/2;
  PX(vsub_INT)(3, no, n, zeros_size);
  PX(pad_array_with_zeros_inside)(n, zeros_size, zeros_start, 2*howmany, (R*) data);
}


static void truncate(
    INT *n, INT *no,
    INT howmany, C *data
    )
{
  INT trunc_size[3], trunc_start[3];

  for(int t=0; t<3; t++)
    trunc_start[t] = (no[t]+1)/2;
  PX(vsub_INT)(3, n, no, trunc_size);
  PX(truncate_array_inside)(n, trunc_size, trunc_start, 2*howmany, (R*) data);
}


void PX(serfft_execute)(
    serfft_plan serfft)
{
  /* execute all initialized serfft plans */
  for(int p=0; p<2; p++)
    if(serfft->plan[p] != NULL)
      X(execute)(serfft->plan[p]);
}


void PX(serfft_trunc_destroy_plan)(
    serfft_trunc_plan ths)
{
  /* plan was already destroyed or never initialized */
  if(ths==NULL)
    return;

  serfft_destroy_plan(ths->serfft);

  /* free memory */
  X(free)(ths);
  /* serfft=NULL; would be senseless, since we can not change the pointer itself */
}



static void serfft_destroy_plan(
    serfft_plan serfft)
{
  /* plan was already destroyed or never initialized */
  if(serfft==NULL)
    return;

  /* take care of unsuccessful FFTW planing */
  for(int t=0; t<2; t++)
    if(serfft->plan[t] != NULL)
      X(destroy_plan)(serfft->plan[t]);

  /* free memory */
  X(free)(serfft);
  /* serfft=NULL; would be senseless, since we can not change the pointer itself */
}

X(plan) PX(plan_fft_for_last_dims)(
    int rank, const INT *n, INT howmany, const INT *local_ni,
    C *data_in, C *data_out, int sign, unsigned fftw_flags
    )
{
  int howmany_rank=1;
  fftw_iodim64 *dims=NULL, howmany_dims[1];
  X(plan) fftplan;
  
  if(rank == 3)
    return NULL;
  
  dims = (fftw_iodim64*) fftw_malloc(sizeof(fftw_iodim64)*(rank-3));
  
  for(int t=3; t<rank; t++)
    dims[t-3].n = n[t];
  
  dims[rank-4].is = howmany;
  for(int t=rank-5; t>2; t--)
    dims[t].is = dims[t+1].is * dims[t+1].n;
  
  for(int t=0; t<rank-3; t++)
    dims[t].os = dims[t].is;
  
  howmany_dims[0].n = local_ni[0]*local_ni[1]*local_ni[2];
  
  howmany_dims[0].is = howmany;
  for(int t=3; t<rank; t++)
    howmany_dims[0].is *= n[t];
  howmany_dims[0].os = howmany_dims[0].is;
  
  fftplan = fftw_plan_guru64_dft(rank-3, dims, howmany_rank, howmany_dims,
        data_in, data_out, sign, fftw_flags);
  fftw_free(dims);
  
  return fftplan;
}

serfft_trunc_plan PX(plan_dft_trunc_1d)(
    const INT *local_ni, const INT *local_n, const INT *local_no,
    INT howmany, C *data_in, C *data_out,
    int sign, unsigned transpose_flags, unsigned fftw_flags,
    PX(truncation) *trc, PX(permutation) *perm
    )
{
  serfft_trunc_plan plan = serfft_trunc_mkplan();

  for(int t=0; t<3; t++){
    plan->local_ni_embed[t] = local_ni[perm->n[t]];
    plan->local_n_embed[t]  = local_n[perm->n[t]];
  }
  
  plan->howmany = howmany;
  plan->data_in = data_in;
  plan->data_out = data_out;
  
  /* plan 1D-FFT */
  plan->serfft = plan_dft_1d(
      local_n, howmany, data_in, data_out, sign, transpose_flags, fftw_flags,
      perm);
  PX(mark_performed_fft)(perm, trc);

  for(int t=0; t<3; t++){
    plan->local_n_trunc[t]  = local_n[perm->n[t]];
    plan->local_no_trunc[t] = local_no[perm->n[t]];
  }
  
  if(plan->serfft == NULL){
    X(free)(plan);
    return NULL;
  }

  return plan;
}


static serfft_trunc_plan serfft_trunc_mkplan(
    void)
{
  serfft_trunc_plan ths =
      (serfft_trunc_plan ) X(malloc)(sizeof(serfft_trunc_plan_s));

  /* initialize to NULL for easy checks */
  ths->serfft=NULL;

  return ths;
}


static serfft_plan plan_dft_1d(
    const INT *local_n, INT howmany, C *data_in, C *data_out,
    int sign, unsigned transpose_flags, unsigned fftw_flags,
    PX(permutation) *perm
    )
{
  int dim_fft;
  INT n[3], is[3], os[3];

  dim_fft = perm->n[PX(get_local_dim)(perm)];
  
  /* input data sizes in row major order (not transposed) */
  for(int p=0; p<3; p++)
    n[p] = local_n[p];
  
  /* input strides in row major order (not transposed) */
  is[perm->n[0]] = local_n[perm->n[1]]*local_n[perm->n[2]];
  is[perm->n[1]] = local_n[perm->n[2]];
  is[perm->n[2]] = 1;

  /* permute if transposed */
  PX(permute_fft_1d)(transpose_flags, perm);

  /* output strides in row major order (transposed) */
  os[perm->n[0]] = local_n[perm->n[1]]*local_n[perm->n[2]];
  os[perm->n[1]] = local_n[perm->n[2]];
  os[perm->n[2]] = 1;

  return plan_serfft(n, is, os, dim_fft, howmany, data_in, data_out, sign, fftw_flags);
}


static serfft_plan plan_serfft(
    INT *n_in, INT *is_in, INT *os_in, int dim_fft, INT howmany, C *data_in, C *data_out,
    int sign, unsigned fftw_flags)
{
  /* FFTW sometimes produces very slow plans for 1dFFT with different is and os.
   * Therefore we replace the FFTW 1dFFT by two FFTW plans, one 1dFFT and one remap.
   * We can choose which one of these plans comes first, which one works out-of-place
   * and which one has different is and os. This results in at most 8 different
   * plans which we compare by time. */

  int p, verbose=0;
  unsigned flag0, flag1, algorithm_flag;
  double time0, time1;
  INT n[3], is[3], os[3];
  serfft_plan serfft0=NULL, serfft1=NULL;
  
  /* strides of 1dFFT must be first for serfft_wrapper*/
  for(p=0; p<3; p++){
    n[p]  = n_in[(dim_fft+p)%3];
    is[p] = is_in[(dim_fft+p)%3];
    os[p] = os_in[(dim_fft+p)%3];
  }
  
  /* do out-of-place plan followed by in-place plan */
  serfft0 = plan_serfft_p(n, is, os, howmany, data_in, data_out, data_out, data_out,
                        sign, fftw_flags, &time0, &flag0);

  /* this plan only differs for out of place FFTs */
  if((data_in != data_out) && (~fftw_flags & FFTW_ESTIMATE)){
    /* do in-place plan followed by out-of-place plan */
    serfft1 = plan_serfft_p(n, is, os, howmany, data_in, data_in, data_in, data_out,
                          sign, fftw_flags, &time1, &flag1);
  }
  else
    time1 = time0;

  /* note result of the comparison */
  algorithm_flag = (time1 < time0) ? flag1 | INPLACE_ON_1ST_REMAP : flag0 | INPLACE_ON_2ND_REMAP;
  if(verbose)
    print_algorithm_flag(algorithm_flag);
  
  /* choose best plan, delete the other one */
  if(time1 < time0){
    serfft_destroy_plan(serfft0);
    return serfft1;
  }
  else{
    serfft_destroy_plan(serfft1);
    return serfft0;
  }
}


static serfft_plan plan_serfft_p(
    INT *n, INT *is, INT *os, INT howmany,
    C *data_in0, C *data_out0, C *data_in1, C *data_out1,
    int sign, unsigned fftw_flags, double *time, unsigned *algorithm_flag)
{
  /* already determined which plan works out of place */

  unsigned flag0, flag1;
  double time1;
  serfft_plan serfft0=NULL, serfft1=NULL;
  
  /* change stride during 1st plan */
  serfft0 = plan_serfft_ps(n, is, os, os, os, howmany, data_in0, data_out0, data_in1, data_out1,
                         sign, fftw_flags, time, &flag0);
  
  /* try this plan only for altered strides */
  if( ((is[0]!=os[0]) || (is[1]!=os[1]) || (is[2]!=os[2])) && (~fftw_flags & FFTW_ESTIMATE)) {
    /* change stride during 2nd plan */
    serfft1 = plan_serfft_ps(n, is, is, is, os, howmany, data_in0, data_out0, data_in1, data_out1,
                           sign, fftw_flags, &time1, &flag1);
  }
  else
    time1 = *time;
  
  /* note result of the comparison */
  *algorithm_flag = (time1 < *time) ? flag1 | STRIDES_CHANGE_ON_2ND_REMAP : flag0 | STRIDES_CHANGE_ON_1ST_REMAP;
  
  /* choose best plan, delete the other one */
  if(time1 < *time){
    *time = time1;
    serfft_destroy_plan(serfft0);
    return serfft1;
  }
  else{
    serfft_destroy_plan(serfft1);
    return serfft0;
  }
}


static serfft_plan plan_serfft_ps(
    INT *n, INT *is0, INT *os0, INT *is1, INT *os1, INT howmany,
    C *data_in0, C *data_out0, C *data_in1, C *data_out1,
    int sign, unsigned fftw_flags, double *time, unsigned *algorithm_flag)
{
  /* already determined which plan works out of place, changes strides */

  double time0=1e10, time1=1e10;
  serfft_plan serfft0=NULL, serfft1=NULL;
  
  /* perform FFT first, remap second */
  serfft0 = serfft_mkplan();
  serfft0->plan[0] = plan_serfft_wrapper(n, is0, os0, howmany, data_in0, data_out0, sign, fftw_flags);
  serfft0->plan[1] = plan_remap_wrapper(n, is1, os1, howmany, data_in1, data_out1, fftw_flags);
  
  /* this plan differs for altered strides or out-of-place */
  serfft1 = serfft_mkplan();
  if( ((is0[0]!=os1[0]) || (is0[1]!=os1[1]) || (is0[2]!=os1[2]) || (data_in0 != data_out1) )
      && (~fftw_flags & FFTW_ESTIMATE)) {
    /* perform remap first, FFT second */
    serfft1->plan[0] = plan_remap_wrapper(n, is0, os0, howmany, data_in0, data_out0, fftw_flags);
    serfft1->plan[1] = plan_serfft_wrapper(n, is1, os1, howmany, data_in1, data_out1, sign, fftw_flags);
  }
  
  time0 = measure_time(serfft0->plan[0])+measure_time(serfft0->plan[1]);
  time1 = measure_time(serfft1->plan[0])+measure_time(serfft1->plan[1]);
  
  *algorithm_flag = (time1 < time0) ? FFT_ON_2ND_REMAP : FFT_ON_1ST_REMAP;
  
  if(time1 < time0){
    *time = time1;
    serfft_destroy_plan(serfft0);
    /* make sure to give back a NULL pointer for broken plans */
    if( (serfft1->plan[0] == NULL) || (serfft1->plan[1] == NULL) ){
      serfft_destroy_plan(serfft1);
      serfft1 = NULL;
    }
    return serfft1;
  }
  else{
    *time = time0;
    /* make sure to give back a NULL pointer for broken plans */
    serfft_destroy_plan(serfft1);
    if( (serfft0->plan[0] == NULL) || (serfft0->plan[1] == NULL) ){
      serfft_destroy_plan(serfft0);
      serfft0 = NULL;
    }
    return serfft0;
  }
}


static void print_algorithm_flag(
    unsigned flag)
{
  static int cycle=1;

  if(flag & INPLACE_ON_2ND_REMAP)
    PX(printf)(MPI_COMM_WORLD, "FFT-Nr. %d, Standard: INPLACE_ON_2ND_REMAP\n", cycle);
  else if(flag & INPLACE_ON_1ST_REMAP)
    PX(printf)(MPI_COMM_WORLD, "FFT-Nr. %d, Non-Standard: INPLACE_ON_1ST_REMAP\n", cycle);
  if(flag & STRIDES_CHANGE_ON_1ST_REMAP)
    PX(printf)(MPI_COMM_WORLD, "FFT-Nr. %d, Standard: STRIDES_CHANGE_ON_1ST_REMAP\n", cycle);
  else if(flag & STRIDES_CHANGE_ON_2ND_REMAP)
    PX(printf)(MPI_COMM_WORLD, "FFT-Nr. %d, Non-Standard: STRIDES_CHANGE_ON_2ND_REMAP\n", cycle);
  if(flag & FFT_ON_1ST_REMAP)
    PX(printf)(MPI_COMM_WORLD, "FFT-Nr. %d, Standard: FFT_ON_1ST_REMAP\n", cycle);
  else if(flag & FFT_ON_2ND_REMAP)
    PX(printf)(MPI_COMM_WORLD, "FFT-Nr. %d, Non-Standard: FFT_ON_2ND_REMAP\n", cycle);
  cycle++;
}


static double measure_time(
    X(plan) plan)
{
  double time=0;
  
  /* 1st execution is often very slow, discard it */
  if(plan != NULL)
    X(execute)(plan);
  
  /* measure 2nd execution */
  /* some in place remaps are not possible and return NULL pointer */
  if(plan != NULL){
    time = -MPI_Wtime();
    X(execute)(plan);
    time += MPI_Wtime();
  }
  else
    time = 1e10;
  /* set time to about 317 years if plan does not exist */
  /* hope we can find a plan that works faster */
  /* otherwise a NULL serfft plan will be returned at the end */

  return time;
}


static serfft_plan serfft_mkplan(
    void)
{
  int p;
  serfft_plan serfft = (serfft_plan) X(malloc)(sizeof(serfft_plan_s));
  
  /* initialize to NULL for easy checks */
  for(p=0; p<2; p++)
    serfft->plan[p]=NULL;
  
  return serfft;
}

static X(plan) plan_serfft_wrapper(
    INT *n, INT *is, INT *os, INT howmany, C *data_in, C *data_out, int sign, unsigned fftw_flags
    )
{
  /* wrap FFTWs guru64 FFT planing, calculate FFT along 1st dim */
  X(iodim64) dims[4];
  
  fill_dims(n, is, os, howmany, dims);
  return X(plan_guru64_dft)(1, dims, 3, &dims[1], data_in, data_out, sign, fftw_flags);
}


static X(plan) plan_remap_wrapper(
    INT *n, INT *is, INT *os, INT howmany, C *data_in, C *data_out, unsigned fftw_flags
    )
{
  /* wrap FFTWs guru64 FFT planing, only reorder without FFT */
  int sign=1;
  X(iodim64) dims[4];
  
  fill_dims(n, is, os, howmany, dims);
  return X(plan_guru64_dft)(0, NULL, 4, dims, data_in, data_out, sign, fftw_flags);
}

static void fill_dims(
    INT *n, INT *is, INT *os, INT howmany,
    X(iodim64) *dims
    )
{
  int p;
  
  for(p=0; p<3; p++){
    dims[p].n  = n[p];
    dims[p].is = is[p]*howmany;
    dims[p].os = os[p]*howmany;
  }
  dims[3].n = howmany;
  dims[3].is = dims[3].os = 1;
}
