/*
 * Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pfft.h"
#include "ipfft.h"
#include <string.h>     /* strcmp */


void PX(get_gctimers)(
    PX(gcplan) ths,
    PX(gctimer) *gctimer_exg, PX(gctimer) *gctimer_red
    )
{
  PX(copy_gctimer)(&ths->timer_exg, gctimer_exg);
  PX(copy_gctimer)(&ths->timer_red, gctimer_red);
}

void PX(convert_gctimer2vec)(
    const PX(gctimer) *gctimer,
    double *times
    )
{
  times[0] = (double) gctimer->iter;
  times[1] = gctimer->whole;
  times[2] = gctimer->pad_zeros;
  times[3] = gctimer->exchange;
}


void PX(convert_vec2gctimer)(
    const double *times,
    PX(gctimer) *gctimer
    )
{
  gctimer->iter      = round(times[0]);
  gctimer->whole     = times[1];
  gctimer->pad_zeros = times[2];
  gctimer->exchange  = times[3];
}


void PX(average_gctimer)(
     PX(gctimer) *gctimer
     )
{
  double times[4];
  
  PX(convert_gctimer2vec)(gctimer, times);
  if(times[0] != 0){
    for(int t=1; t<4; t++)
      times[t] /= times[0];
    times[0] = 1;
  }
  PX(convert_vec2gctimer)(times, gctimer);
}

void PX(copy_gctimer)(
    const PX(gctimer) *orig,
    PX(gctimer) *copy
    )
{
  double times[4];
  
  PX(convert_gctimer2vec)(orig, times);
  PX(convert_vec2gctimer)(times, copy);
}

void PX(reduce_max_gctimer)(
    MPI_Comm comm, const PX(gctimer) *gctimer,
    PX(gctimer) *gctimer_max
    )
{
  double times[4], times_max[4];
  
  PX(convert_gctimer2vec)(gctimer, times);
  MPI_Reduce(times, times_max, 4, MPI_DOUBLE, MPI_MAX, 0, comm);
  PX(convert_vec2gctimer)(times_max, gctimer_max);
}

void PX(add_gctimers)(
    const PX(gctimer) *sum1, const PX(gctimer) *sum2,
    PX(gctimer) *res
    )
{
  double t1[4], t2[4], sum[4];
  
  PX(convert_gctimer2vec)(sum1, t1);
  PX(convert_gctimer2vec)(sum2, t2);
  for(int t=0; t<4; t++)
    sum[t] = t1[t] + t2[t];
  PX(convert_vec2gctimer)(sum, res);
}


void PX(reset_gctimer)(
    PX(gctimer) *gctimer
    )
{
  double times[4];
  
  for(int t=0; t<4; t++)
    times[t] = 0;
  PX(convert_vec2gctimer)(times, gctimer);
}

void PX(reset_gctimers)(
    PX(gcplan) ths
    )
{
  PX(reset_gctimer)(&ths->timer_exg);
  PX(reset_gctimer)(&ths->timer_red);
}


static int file_exists(
    const char *restrict name
    )
{
  FILE *f;
  
  f=fopen(name, "r");
  if(f != NULL)
    fclose(f);
  
  return (f != NULL);
}

static FILE* open_or_create_file_to_append(
    MPI_Comm comm, const char *restrict name
    )
{
  FILE *f=fopen(name, "a+");
  if(f==NULL){
    PX(fprintf)(comm, stderr, "Error: Cannot open file %s.\n", name);
    exit(1);
  }
  return f;
}

static void write_info_header(
    MPI_Comm comm, FILE *restrict file
    )
{
  PX(fprintf)(comm, file, "\n%% n - FFT size\n");
  PX(fprintf)(comm, file, "%% np - process grid\n");
  PX(fprintf)(comm, file, "%% procs - number of processes\n");
  PX(fprintf)(comm, file, "%% index(i) = log(procs(i)) + 1\n");
}

static int get_matlab_index(
    MPI_Comm comm
    )
{
  int size;
  MPI_Comm_size(comm, &size);
  return (int)round(log(size)/log(2))+1;
}

static void write_run_specific_infos(
    MPI_Comm comm, FILE *restrict file, PX(gcplan) ths
    )
{
  
  int size;
  int idx = get_matlab_index(comm);
  MPI_Comm_size(comm, &size);
  
  PX(fprintf)(comm, file, "\nindex(%d) = %d;  ", idx, idx);
  PX(fprintf)(comm, file, "procs(%d) = %d;  ", idx, size);
  PX(fprintf)(comm, file, "np_gc(%d, 1:3) = [%d %d %d];\n",
      idx, ths->np[0], ths->np[1], ths->np[2]);
  PX(fprintf)(comm, file, "n_gc(%d, 1:3) = [%td %td %td];  ",
      idx, ths->n[0], ths->n[1], ths->n[2]);
  PX(fprintf)(comm, file, "gc_below(%d, 1:3) = [%td %td %td];  ",
      idx, ths->gc_below[0], ths->gc_below[1], ths->gc_below[2]);
  PX(fprintf)(comm, file, "gc_above(%d, 1:3) = [%td %td %td];\n",
      idx, ths->gc_above[0], ths->gc_above[1], ths->gc_above[2]);
}

static void fprint_average_gctimer_internal(
    MPI_Comm comm, FILE *restrict file, const char *prefix, PX(gctimer) *ths,
    unsigned flags
    )
{
  int idx = get_matlab_index(comm);
  PX(gctimer) mt;
  PX(reduce_max_gctimer)(comm, ths, &mt);
  PX(average_gctimer)(&mt);
  
  if(flags & PFFT_PRINT_TIMER_BASIC){
    PX(fprintf)(comm, file, "%s_gc_iter(%d)      = %d;  ", prefix, idx, ths->iter);
    PX(fprintf)(comm, file, "%s_gcells(%d)       = %.3e;\n", prefix, idx, mt.whole);
  } else if(flags & PFFT_PRINT_TIMER_ADV){
    PX(fprintf)(comm, file, "%s_gc_pad_zeros(%d) = %.3e;  ", prefix, idx, mt.pad_zeros);
    PX(fprintf)(comm, file, "%s_gc_exchange(%d)  = %.3e;\n", prefix, idx, mt.exchange);
  }
}

static void fprint_average_gctimer(
    MPI_Comm comm, FILE *restrict file, PX(gcplan) ths, unsigned flags
    )
{
  fprint_average_gctimer_internal(comm, file, "gcells_exg", &ths->timer_exg, flags);
  fprint_average_gctimer_internal(comm, file, "gcells_red", &ths->timer_red, flags);
}

void PX(write_average_gctimer)(
    MPI_Comm comm, PX(gcplan) ths, const char *restrict name
    )
{
  int newfile;
  FILE *f;
  
  newfile = !file_exists(name);
  f = open_or_create_file_to_append(comm, name);
  if( newfile )
    write_info_header(comm, f);
  
  write_run_specific_infos(comm, f, ths);
  fprint_average_gctimer(comm, f, ths, PFFT_PRINT_TIMER_BASIC);
  
  fclose(f);
}

void PX(write_average_gctimer_adv)(
    MPI_Comm comm, PX(gcplan) ths, const char *restrict name
    )
{
  FILE *f;
  PX(write_average_gctimer)(comm, ths, name);
  f = open_or_create_file_to_append(comm, name);
  fprint_average_gctimer(comm, f, ths, PFFT_PRINT_TIMER_ADV);
  fclose(f);
}

void PX(print_average_gctimer)(
    MPI_Comm comm, PX(gcplan) ths
    )
{
  write_info_header(comm, stdout);
  write_run_specific_infos(comm, stdout, ths);
  fprint_average_gctimer(comm, stdout, ths, PFFT_PRINT_TIMER_BASIC);
}

void PX(print_average_gctimer_adv)(
    MPI_Comm comm, PX(gcplan) ths
    )
{
  PX(print_average_gctimer)(comm, ths);
  fprint_average_gctimer(comm, stdout, ths, PFFT_PRINT_TIMER_ADV);
}
