/*
 * Copyright (c) 2003, 2007-8 Matteo Frigo
 * Copyright (c) 2003, 2007-8 Massachusetts Institute of Technology
 * Copyright (c) 2010 Michael Pippig
 *
 * The following statement of license applies *only* to this header file,
 * and *not* to the other files distributed with FFTW or derived therefrom:
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef PFFT_H
#define PFFT_H 1

#include "fftw3.h"
#include <mpi.h>
#include <stdarg.h>     /* va_list */

/* move following define into an internal header */
#ifdef __cplusplus
#  define BEGIN_C_DECLS extern "C" {
#  define END_C_DECLS   }
#else /* !__cplusplus */
#  define BEGIN_C_DECLS
#  define END_C_DECLS
#endif /* __cplusplus */

BEGIN_C_DECLS

#define PFFT_EXTERN extern
#define PFFT_CONCAT(prefix, name) prefix ## name


/*
  huge second-order macro that defines prototypes for all API
  functions.  We expand this macro for each supported precision

  PX: PFFT name-mangling macro (parallel)
  X : FFTW name-mangling macro (serial)
  R : real data type
  C : complex data type
*/
#define PFFT_DEFINE_API(PX, X, R, C, INT)                                               \
                                                                                        \
  typedef struct PX(plan_s) *PX(plan);                                                  \
  typedef struct PX(gcplan_s) *PX(gcplan);                                              \
  typedef struct{                                                                       \
    INT ni[3];                                                                          \
    INT no[3];                                                                          \
    int is_transformed[3];                                                              \
  } PX(truncation);                                                                     \
  typedef struct{                                                                       \
    int n[3];                                                                           \
    int np[3];                                                                          \
  } PX(permutation);                                                                    \
  typedef struct {                                                                      \
    unsigned fft_flag[3];                                                               \
    int first_dims[3];                                                                  \
    unsigned remap_flag[3];                                                             \
  } PX(profile);                                                                        \
  typedef struct {                                                                      \
    int iter;                                                                           \
    double whole;                                                                       \
    double fft[4];                                                                      \
    double remap[3];                                                                    \
  } PX(timer);                                                                          \
  typedef struct {                                                                      \
    int iter;                                                                           \
    double whole;                                                                       \
    double pad_zeros;                                                                   \
    double exchange;                                                                    \
  } PX(gctimer);                                                                        \
                                                                                        \
  PFFT_EXTERN void PX(init)(void);                                                      \
  PFFT_EXTERN void PX(cleanup)(void);                                                   \
                                                                                        \
  PFFT_EXTERN void PX(execute)(PX(plan) fftplan);                                       \
  PFFT_EXTERN void PX(destroy_plan)(PX(plan) fftplan);                                  \
                                                                                        \
  PFFT_EXTERN int PX(create_procmesh_2d)(                                               \
      MPI_Comm comm, int np0, int np1,                                                  \
      MPI_Comm *comm_cart_2d);                                                          \
                                                                                        \
  PFFT_EXTERN void PX(init_input_3d)(                                                   \
      const INT *n, const INT *local_n, const INT *local_n_start,                       \
      MPI_Comm comm, C *data);                                                          \
  PFFT_EXTERN void PX(init_many_input_3d)(                                              \
      const INT *n, const INT *ni, INT howmany,                                         \
      const INT *local_n, const INT *local_n_start, MPI_Comm comm,                      \
      C *data);                                                                         \
  PFFT_EXTERN void PX(check_output_3d)(                                                 \
      C *data, const INT *n,                                                            \
      const INT *local_n, const INT *local_n_start,                                     \
      MPI_Comm comm, int iterations, int iscale);                                       \
  PFFT_EXTERN void PX(check_many_output_3d)(                                            \
      C *data, const INT *n, const INT *ni, INT howmany,                                \
      const INT *local_ni, const INT *local_i_start,                                    \
      MPI_Comm comm, int iterations, int iscale);                                       \
                                                                                        \
  PFFT_EXTERN int PX(check_3d)(                                                         \
      const INT *n, MPI_Comm comm_cart_2d,                                              \
      unsigned pfft_flags, int verbose);                                                \
  PFFT_EXTERN int PX(check_many_3d)(                                                    \
      const INT *n, const INT *ni, const INT *no,                                       \
      const INT *block_col, const INT *block_row,                                       \
      MPI_Comm comm_cart_2d, unsigned pfft_flags, int verbose);                         \
  PFFT_EXTERN int PX(check_guru_3d)(                                                    \
      const INT *n, const INT *ni, const INT *no,                                       \
      const PX(profile) *prf, const PX(permutation) *perm,                              \
    const INT *block_col, const INT *block_row,                                         \
    MPI_Comm comm_cart_2d, int verbose);                                                \
                                                                                        \
  PFFT_EXTERN INT PX(local_size_3d)(                                                    \
      const INT *n, MPI_Comm comm_cart_2d, unsigned pfft_flags,                         \
      INT *local_ni, INT *local_i_start,                                                \
      INT *local_no, INT *local_o_start);                                               \
  PFFT_EXTERN INT PX(local_size)(                                                       \
      int rank, const INT *n, MPI_Comm comm_cart_2d, unsigned pfft_flags,               \
      INT *local_ni, INT *local_i_start,                                                \
      INT *local_no, INT *local_o_start);                                               \
  PFFT_EXTERN INT PX(local_size_many)(                                                  \
      int rank, const INT *n, const INT *ni, const INT *no,                             \
      INT howmany, const INT *block_col, const INT *block_row,                          \
      MPI_Comm comm_cart_2d, unsigned pfft_flags,                                       \
      INT *local_ni, INT *local_i_start,                                                \
      INT *local_no, INT *local_o_start);                                               \
  PFFT_EXTERN INT PX(local_size_guru)(                                                  \
      int rank, const INT *n, const INT *ni, const INT *no,                             \
      const PX(profile) *prf, const PX(permutation) *perm,                              \
      INT howmany, const INT *block_col, const INT *block_row,                          \
      MPI_Comm comm_cart_2d,                                                            \
      INT *local_ni, INT *local_i_start,                                                \
      INT *local_no, INT *local_o_start,                                                \
      PX(permutation)* perm_out);                                                       \
                                                                                        \
  PFFT_EXTERN PX(plan) PX(plan_dft_3d)(                                                 \
      const INT *n, C *data_in, C *data_out,                                            \
      MPI_Comm comm_cart_2d,                                                            \
      int sign, unsigned pfft_flags, unsigned fftw_flags);                              \
  PFFT_EXTERN PX(plan) PX(plan_dft)(                                                    \
      int rank, const INT *n, C *data_in, C *data_out,                                  \
      MPI_Comm comm_cart_2d,                                                            \
      int sign, unsigned pfft_flags, unsigned fftw_flags);                              \
  PFFT_EXTERN PX(plan) PX(plan_many_dft)(                                               \
      int rank, const INT *n, const INT *ni, const INT *no,                             \
      INT howmany, const INT *block_col, const INT *block_row,                          \
      C *data_in, C *data_out, MPI_Comm comm_cart_2d,                                   \
      int sign, unsigned pfft_flags, unsigned fftw_flags);                              \
  PFFT_EXTERN PX(plan) PX(plan_guru_dft)(                                               \
      int rank, const INT *n, const INT *ni, const INT *no,                             \
      const PX(profile) *prf, const PX(permutation) *perm,                              \
      INT howmany, const INT *block_col, const INT *block_row,                          \
      C *data_in, C *data_out, MPI_Comm comm_cart_2d,                                   \
      int sign, unsigned fftw_flags);                                                   \
                                                                                        \
  PFFT_EXTERN void PX(init_default_profile)(                                            \
      unsigned pfft_flags,                                                              \
      PX(profile) *prf, PX(permutation) *perm);                                         \
                                                                                        \
  /*  n0/np0 x n1/np1 x n2  to  N0/np1 x N1/np0 x N2  */                                \
  PFFT_EXTERN void PX(profile_n012_np102)(PX(profile) *prf);                            \
  /*  n0/np0 x n1/np1 x n2  to  N0/np1 x N2/np0 x N1  */                                \
  PFFT_EXTERN void PX(profile_n021_np102)(PX(profile) *prf);                            \
  /*  n0/np0 x n1/np1 x n2  to  N2/np0 x N0/np1 x N1  */                                \
  PFFT_EXTERN void PX(profile_n201_np012)(PX(profile) *prf);                            \
  /*  n0/np0 x n1/np1 x n2  to  N1/np0 x N2/np1 x N0  */                                \
  PFFT_EXTERN void PX(profile_n120_np012)(PX(profile) *prf);                            \
  /*  n0/np0 x n1/np1 x n2  to  N1/np0 x N2/np1 x N0  */                                \
  PFFT_EXTERN void PX(profile_n120_np012_IN)(PX(profile) *prf);                         \
  /*  n0/np0 x n1/np1 x n2  to  N2/np0 x N0/np1 x N1  */                                \
  PFFT_EXTERN void PX(profile_n201_np012_IN)(PX(profile) *prf);                         \
  PFFT_EXTERN void PX(profile_n201_np012_IN1)(PX(profile) *prf);                        \
  PFFT_EXTERN void PX(profile_n201_np012_IN2)(PX(profile) *prf);                        \
  /*  n0/np0 x n1/np1 x n2  to  N0/np1 x N2/np0 x N1  */                                \
  /*  self inverse  */                                                                  \
  PFFT_EXTERN void PX(profile_n210_np102)(PX(profile) *prf);                            \
  /*  n0/np0 x n1/np1 x n2  to  N0/np1 x N2/np0 x N1  */                                \
  /*  self inverse, out of place only  */                                               \
  PFFT_EXTERN void PX(profile_n021_np102_outofplace)(PX(profile) *prf);                 \
  /*  n0/np0 x n1/np1 x n2  to  N0/np1 x N2/np0 x N1  */                                \
  /*  self inverse  */                                                                  \
  PFFT_EXTERN void PX(profile_n021_np102_inplace)(PX(profile) *prf);                    \
                                                                                        \
  PFFT_EXTERN INT PX(prod_INT)(                                                         \
      int d, const INT *vec);                                                           \
  PFFT_EXTERN int PX(equal_INT)(                                                        \
      int d, const INT *vec1, const INT *vec2);                                         \
  PFFT_EXTERN void PX(vcopy_INT)(                                                       \
      int d, const INT *vec1,                                                           \
      INT *vec2);                                                                       \
  PFFT_EXTERN void PX(vadd_INT)(                                                        \
      int d, const INT *vec1, const INT *vec2,                                          \
      INT *sum);                                                                        \
  PFFT_EXTERN void PX(vsub_INT)(                                                        \
      int d, const INT *vec1, const INT *vec2,                                          \
      INT *sum);                                                                        \
                                                                                        \
  PFFT_EXTERN void *PX(malloc)(                                                         \
      size_t n, MPI_Comm comm);                                                         \
  PFFT_EXTERN void PX(free)(                                                            \
      void *p);                                                                         \
                                                                                        \
  PFFT_EXTERN void PX(get_arg_INT)(                                                     \
      int argc, char **argv, char *name, int neededArgs, INT *parameter);               \
  PFFT_EXTERN void PX(get_arg_int)(                                                     \
      int argc, char **argv, char *name, int neededArgs, int* parameter);               \
                                                                                        \
                                                                                        \
                                                                                        \
  PFFT_EXTERN INT PX(local_size_gcells_3d)(                                             \
    const INT *local_n, const INT *local_start, INT alloc_local,                        \
    const INT *num_gcells_below, const INT *num_gcells_above,                           \
    INT *local_ngc, INT *local_gc_start);                                               \
  PFFT_EXTERN INT PX(local_size_gcells)(                                                \
    int rank, const INT *local_n, const INT *local_start, INT alloc_local,              \
    const INT *num_gcells_below, const INT *num_gcells_above,                           \
    INT *local_ngc, INT *local_gc_start);                                               \
  PFFT_EXTERN INT PX(local_size_many_gcells)(                                           \
    int rank, const INT *local_n, const INT *local_start, INT alloc_local,              \
    INT howmany, const INT *num_gcells_below, const INT *num_gcells_above,              \
    INT *local_ngc, INT *local_gc_start);                                               \
  PFFT_EXTERN INT PX(local_size_gcells_guru)(                                           \
    int rank, const INT *local_n, const INT *local_start, INT alloc_local,              \
    INT howmany, const INT *num_gcells_below, const INT *num_gcells_above,              \
    INT *local_ngc, INT *local_gc_start);                                               \
                                                                                        \
  PFFT_EXTERN PX(gcplan) PX(plan_gcells_3d)(                                            \
    const INT *n, const INT *num_gcells_below, const INT *num_gcells_above,             \
    C *data, MPI_Comm comm_cart_2d, unsigned pfft_flags);                               \
  PFFT_EXTERN PX(gcplan) PX(plan_gcells)(                                               \
    int rank, const INT *n, const INT *num_gcells_below, const INT *num_gcells_above,   \
    C *data, MPI_Comm comm_cart_2d, unsigned pfft_flags);                               \
  PFFT_EXTERN PX(gcplan) PX(plan_many_gcells)(                                          \
    int rank, const INT *n, INT howmany, const INT *block_col, const INT *block_row,    \
    const INT *num_gcells_below, const INT *num_gcells_above,                           \
    C *data, MPI_Comm comm_cart_2d, unsigned pfft_flag);                                \
  PFFT_EXTERN PX(gcplan) PX(plan_gcells_guru)(                                          \
    int rank, const INT *n, const PX(permutation) *perm,                                \
    INT howmany, const INT *block_col, const INT *block_row,                            \
    const INT *num_gcells_below, const INT *num_gcells_above,                           \
    C *data, MPI_Comm comm_cart_2d, unsigned alg_flags);                                \
                                                                                        \
  PFFT_EXTERN void PX(destroy_gcplan)(PX(gcplan) ths);                                  \
  PFFT_EXTERN void PX(exchange)(PX(gcplan) ths);                                        \
  PFFT_EXTERN void PX(reduce)(PX(gcplan) ths);                                          \
                                                                                        \
  void PX(apr_complex_3d)(                                                              \
      const fftw_complex *data, const INT *local_n,                                     \
      const INT *local_n_start, const char *name, MPI_Comm comm);                       \
  void PX(apr_complex_permuted_3d)(                                                     \
      const fftw_complex *data, const INT *local_n,                                     \
      const INT *local_n_start,                                                         \
      int perm0, int perm1, int perm2, const char *name, MPI_Comm comm);                \
                                                                                        \
  PFFT_EXTERN void PX(get_args)(                                                        \
      int argc, char **argv, const char *name,                                          \
      int neededArgs, unsigned type,                                                    \
      void *parameter);                                                                 \
                                                                                        \
  PFFT_EXTERN void PX(convert_timer2vec)(                                               \
      const PX(timer) *timer, double *times);                                           \
  PFFT_EXTERN void PX(convert_vec2timer)(                                               \
    const double *times, PX(timer) *timer);                                             \
  PFFT_EXTERN void PX(average_timer)(                                                   \
    PX(timer) *timer);                                                                  \
  PFFT_EXTERN void PX(copy_timer)(                                                      \
    const PX(timer) *orig, PX(timer) *copy);                                            \
  PFFT_EXTERN void PX(reduce_max_timer)(                                                \
    MPI_Comm comm, const PX(timer) *timer, PX(timer) *timer_max);                       \
  PFFT_EXTERN void PX(add_timers)(                                                      \
    const PX(timer) *sum1, const PX(timer) *sum2, PX(timer) *res);                      \
                                                                                        \
  PFFT_EXTERN void PX(get_timer)(                                                       \
    PX(plan) ths, PX(timer) *timer);                                                    \
  PFFT_EXTERN void PX(reset_timer)(                                                     \
      PX(plan) ths);                                                                    \
  PFFT_EXTERN void PX(print_average_timer)(                                             \
      MPI_Comm comm, PX(plan) ths);                                                     \
  PFFT_EXTERN void PX(print_average_timer_adv)(                                         \
      MPI_Comm comm, PX(plan) ths);                                                     \
  PFFT_EXTERN void PX(write_average_timer)(                                             \
      MPI_Comm comm, PX(plan) ths, const char *restrict name);                          \
  PFFT_EXTERN void PX(write_average_timer_adv)(                                         \
      MPI_Comm comm, PX(plan) ths, const char *restrict name);                          \
                                                                                        \
  PFFT_EXTERN void PX(convert_gctimer2vec)(                                             \
      const PX(gctimer) *gctimer, double *times);                                       \
  PFFT_EXTERN void PX(convert_vec2gctimer)(                                             \
    const double *times, PX(gctimer) *gctimer);                                         \
  PFFT_EXTERN void PX(average_gctimer)(                                                 \
    PX(gctimer) *gctimer);                                                              \
  PFFT_EXTERN void PX(copy_gctimer)(                                                    \
    const PX(gctimer) *orig, PX(gctimer) *copy);                                        \
  PFFT_EXTERN void PX(reduce_max_gctimer)(                                              \
    MPI_Comm comm, const PX(gctimer) *gctimer, PX(gctimer) *gctimer_max);               \
  PFFT_EXTERN void PX(add_gctimers)(                                                    \
    const PX(gctimer) *sum1, const PX(gctimer) *sum2, PX(gctimer) *res);                \
                                                                                        \
  PFFT_EXTERN void PX(get_gctimers)(                                                    \
    PX(gcplan) ths, PX(gctimer) *gctimer_exg, PX(gctimer) *gctimer_red);                \
  PFFT_EXTERN void PX(reset_gctimers)(                                                  \
      PX(gcplan) ths);                                                                  \
  PFFT_EXTERN void PX(print_average_gctimer)(                                           \
      MPI_Comm comm, PX(gcplan) ths);                                                   \
  PFFT_EXTERN void PX(print_average_gctimer_adv)(                                       \
      MPI_Comm comm, PX(gcplan) ths);                                                   \
  PFFT_EXTERN void PX(write_average_gctimer)(                                           \
      MPI_Comm comm, PX(gcplan) ths, const char *restrict name);                        \
  PFFT_EXTERN void PX(write_average_gctimer_adv)(                                       \
      MPI_Comm comm, PX(gcplan) ths, const char *restrict name);                        \
                                                                                        \
  PFFT_EXTERN void PX(vfprintf)(                                                        \
      MPI_Comm comm, FILE *restrict stream, const char *restrict format, va_list ap);   \
  PFFT_EXTERN void PX(fprintf)(                                                         \
      MPI_Comm comm, FILE *restrict stream, const char *restrict format, ...);          \
  PFFT_EXTERN void PX(printf)(                                                          \
      MPI_Comm comm, const char *restrict format, ...);



#define PFFT_MANGLE_DOUBLE(name) PFFT_CONCAT(pfft_, name)
#define PFFT_MANGLE_FLOAT(name) PFFT_CONCAT(pfftf_, name)
#define PFFT_MANGLE_LONG_DOUBLE(name) PFFT_CONCAT(pfftl_, name)

PFFT_DEFINE_API(
    PFFT_MANGLE_DOUBLE, FFTW_MANGLE_DOUBLE,
    double, fftw_complex, ptrdiff_t)
PFFT_DEFINE_API(
    PFFT_MANGLE_FLOAT, FFTW_MANGLE_FLOAT,
    float, fftwf_complex, ptrdiff_t)
PFFT_DEFINE_API(
    PFFT_MANGLE_LONG_DOUBLE, FFTW_MANGLE_LONG_DOUBLE,
    long double, fftwl_complex, ptrdiff_t)



#define PFFT_TRANSPOSE_NONE       (0U)
#define PFFT_TRANSPOSE_012        (0U)
#define PFFT_TRANSPOSE_102        (1U<< 0)
#define PFFT_TRANSPOSE_021        (1U<< 1)
#define PFFT_TRANSPOSE_201        (1U<< 2)
#define PFFT_TRANSPOSE_120        (1U<< 3)

#define PFFT_FORWARD              (0U)
#define PFFT_BACKWARD             (1U<< 4)
#define PFFT_TRANSPOSED           (1U<< 5)
#define PFFT_P3DFFT               (1U<< 6)

#define PFFT_GC_BEFORE_FFT        (0U)
#define PFFT_GC_AFTER_FFT         (1U<< 7)

#define PFFT_DEFAULT_BLOCK        FFTW_MPI_DEFAULT_BLOCK

#define PFFT_DEFAULT_BLOCKS       NULL
#define PFFT_NO_GCELLS            NULL

#define FPFFT_DEFAULT_BLOCKS      (-1)
#define FPFFT_NO_GCELLS           (-1)

#define PFFT_INT                  (1U)
#define PFFT_PTRDIFF_T            (2U)
#define PFFT_FLOAT                (3U)
#define PFFT_DOUBLE               (4U)

END_C_DECLS

#endif /* !PFFT_H */
