/*
 * Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pfft.h"
#include "ipfft.h"
#include <string.h>     /* strcmp */


void PX(permute)(int p0, int p1, int p2, int* q)
{
  q[0] = p0; q[1] = p1; q[2] = p2;
}


void PX(permute_vector)(unsigned fft_flag, int* permutation)
{
  if( PX(flag_active)(fft_flag, PFFT_TRANSPOSE_102) )
    PX(permute)(permutation[1], permutation[0], permutation[2], permutation);
  else if( PX(flag_active)(fft_flag, PFFT_TRANSPOSE_021) )
    PX(permute)(permutation[0], permutation[2], permutation[1], permutation);
  else if( PX(flag_active)(fft_flag, PFFT_TRANSPOSE_201) )
    PX(permute)(permutation[2], permutation[0], permutation[1], permutation);
  else if( PX(flag_active)(fft_flag, PFFT_TRANSPOSE_120) )
    PX(permute)(permutation[1], permutation[2], permutation[0], permutation);
}

void PX(init_truncation)(
    const INT *ni, const INT *no,
    PX(truncation) *trc
    )
{
  for(int t=0; t<3; t++){
    trc->ni[t] = ni[t];
    trc->no[t] = no[t];
    trc->is_transformed[t] = 0;
  }
}

void PX(decompose_fft_trunc_3d)(
    const INT *n, const PX(truncation) *trc, const PX(permutation) *perm,
    const INT *block_col, const INT *block_row,
    MPI_Comm comm_col, MPI_Comm comm_row,
    INT *local_ni, INT *local_n, INT *local_no
    )
{
  int dim_fft_perm;
  INT n_trunc[3], dummy[3];

  dim_fft_perm = perm->n[ PX(get_local_dim)(perm) ];

  PX(get_current_global_size)(trc, n_trunc);

  PX(decompose_3d)(n_trunc, perm, block_col, block_row, comm_col, comm_row,
      local_ni, dummy);

  n_trunc[dim_fft_perm] = n[dim_fft_perm];
  PX(decompose_3d)(n_trunc, perm, block_col, block_row, comm_col, comm_row,
      local_n, dummy);

  n_trunc[dim_fft_perm] = trc->no[dim_fft_perm];
  PX(decompose_3d)(n_trunc, perm, block_col, block_row, comm_col, comm_row,
      local_no, dummy);
}

void PX(get_current_global_size)(
    const PX(truncation) *trc,
    INT *n_trunc
    )
{
  for(int t=0; t<3; t++)
    n_trunc[t] = (trc->is_transformed[t]) ? trc->no[t] : trc->ni[t];
}


void PX(decompose_3d)(
    const INT *n, const PX(permutation) *perm,
    const INT *block_col, const INT *block_row,
    MPI_Comm comm_col, MPI_Comm comm_row,
    INT *local_n, INT *local_n_start
    )
{
  int q, coord_col, coord_row;

  PX(get_mpi_cart_coord)(comm_col, &coord_col);
  PX(get_mpi_cart_coord)(comm_row, &coord_row);

  for(int t=0; t<3; t++){
    q = perm->n[t];
    switch (perm->np[t]){
    case 0:
      PX(decompose_1d)(n[q], block_col[q], coord_col, &local_n[q], &local_n_start[q]); break;
    case 1:
      PX(decompose_1d)(n[q], block_row[q], coord_row, &local_n[q], &local_n_start[q]); break;
    case 2:
      local_n[q] = n[q]; local_n_start[q] = 0;
    }
  }
}


void PX(decompose_1d)(
    INT n, INT block_size, int which_block,
    INT *local_n, INT *local_n_start
    )
{
  PX(local_block_size_and_offset)(n, block_size, which_block, local_n, local_n_start);
}


int PX(create_procmesh_2d)(
    MPI_Comm comm, int np0, int np1,
    MPI_Comm *comm_cart_2d
    )
{
  int periods[2], reorder=1, np[2], num_procs, ret = 0;

  *comm_cart_2d = MPI_COMM_NULL;

  MPI_Comm_size(comm, &num_procs);

  if(num_procs != np0*np1){
    ret = 1;
  } else {
    np[0] = np0; np[1] = np1;
    periods[0] = periods[1] = 1;
    ret = MPI_Cart_create(comm, 2, np, periods, reorder, comm_cart_2d);
  }

  return ret;
}


int PX(is_cart_procmesh_2d)(
    MPI_Comm comm_cart_2d
    )
{
  int status, ndims;

  MPI_Topo_test(comm_cart_2d, &status);
  MPI_Cartdim_get(comm_cart_2d, &ndims);

  return ( (status == MPI_CART) && (ndims == 2) );
}

void PX(split_cart_procmesh_2d)(
    MPI_Comm comm_cart_2d,
    MPI_Comm *comm_col, MPI_Comm* comm_row
    )
{
  int remain_dims[2];

  remain_dims[0] = 1; remain_dims[1] = 0;
  MPI_Cart_sub(comm_cart_2d, remain_dims, comm_col);

  remain_dims[0] = 0; remain_dims[1] = 1;
  MPI_Cart_sub(comm_cart_2d, remain_dims, comm_row);
}


void PX(pad_permuted_array_with_zeros_outside)(
    const INT *orig_array_size, const INT *zeros_below, const INT *zeros_above,
    const PX(permutation) *perm, INT tuple_size, R *real_data
    )
{
  INT orig_array_size_perm[3], zeros_below_perm[3], zeros_above_perm[3];
  
  for(int t=0; t<3; t++){
    orig_array_size_perm[t]   = orig_array_size[perm->n[t]];
    zeros_below_perm[t] = zeros_below[perm->n[t]];
    zeros_above_perm[t] = zeros_above[perm->n[t]];
  }
  
  PX(pad_array_with_zeros_outside)(
      orig_array_size_perm, zeros_below_perm, zeros_above_perm, tuple_size,
      real_data);
}


void PX(truncate_permuted_array_outside)(
    const INT *orig_array_size, const INT *trunc_below, const INT *trunc_above,
    const PX(permutation) *perm, INT tuple_size, R *real_data
    )
{
  INT orig_array_size_perm[3], trunc_below_perm[3], trunc_above_perm[3];

  for(int t=0; t<3; t++){
    orig_array_size_perm[t] = orig_array_size[perm->n[t]];
    trunc_below_perm[t]  = trunc_below[perm->n[t]];
    trunc_above_perm[t] = trunc_above[perm->n[t]];
  }

  PX(truncate_array_outside)(
      orig_array_size_perm, trunc_below_perm, trunc_above_perm, tuple_size,
      real_data);
}


void PX(truncate_permuted_array_inside)(
    const INT *n, const INT *trunc_size, const INT *trunc_start,
    const PX(permutation) *perm, INT tuple_size,
    R *real_data
    )
{
  INT n_perm[3], trunc_size_perm[3], trunc_start_perm[3];
  
  for(int t=0; t<3; t++){
    n_perm[t] = n[perm->n[t]];
    trunc_size_perm[t] = trunc_size[perm->n[t]];
    trunc_start_perm[t] = trunc_start[perm->n[t]];
  }
  
  PX(truncate_array_inside)(
      n_perm, trunc_size_perm, trunc_start_perm, tuple_size,
      real_data);
}



void PX(pad_array_with_zeros_outside)(
    const INT *n, const INT *zeros_below, const INT *zeros_above,
    INT tuple_size, R *real_data
    )
{
  INT m, mo;
  INT n_start[3], n_end[3], no[3];
  INT k0, k1, k2, h;
  int k0_is_outside, k1_is_outside, k2_is_outside, index_is_outside;

  PX(vcopy_INT)(3, zeros_below, n_start);
  PX(vadd_INT)(3, n_start, n, n_end);
  PX(vadd_INT)(3, n_end, zeros_above, no);

  if( PX(equal_INT)(3, n, no) )
    return;

  m  = tuple_size * PX(prod_INT)(3, n) - 1;
  mo = tuple_size * PX(prod_INT)(3, no) - 1;
  for(k0 = no[0] - 1 ; k0 >= 0; k0--){
    k0_is_outside = (k0 < n_start[0]) || (n_end[0] <= k0);
    for(k1 = no[1] - 1 ; k1 >= 0; k1--){
      k1_is_outside = (k1 < n_start[1]) || (n_end[1] <= k1);
      for(k2 = no[2] - 1 ; k2 >= 0; k2--){
        k2_is_outside = (k2 < n_start[2]) || (n_end[2] <= k2);
        index_is_outside = k0_is_outside || k1_is_outside || k2_is_outside;
        for(h = 0; h < tuple_size; h++)
          real_data[mo--] = (index_is_outside) ? 0 : real_data[m--];
      }
    }
  }
}


void PX(truncate_array_outside)(
    const INT *n, const INT *trunc_below, const INT *trunc_above,
    INT tuple_size, R *real_data
    )
{
  INT n_total, mo, k0, k1, k2, h;
  INT no_start[3], no_end[3], no[3];

  PX(vcopy_INT)(3, trunc_below, no_start);
  PX(vsub_INT)(3, n, trunc_above, no_end);
  PX(vsub_INT)(3, no_end, trunc_below, no);

  if( PX(equal_INT)(3, n, no) )
    return;

  mo = 0;
  for(k0 = no_start[0]; k0 < no_end[0]; k0++){
    for(k1 = no_start[1]; k1 < no_end[1]; k1++){
      for(k2 = no_start[2]; k2 < no_end[2]; k2++){
        for(h = 0; h < tuple_size; h++)
          real_data[mo++] = real_data[h+tuple_size*(k2+n[2]*(k1+n[1]*k0))];
      }
    }
  }

  n_total = tuple_size * PX(prod_INT)(3, n);
  for(k0 = mo; k0 < n_total; k0++)
    real_data[k0] = 0;
}


void PX(pad_array_with_zeros_inside)(
    const INT *n, const INT *zeros_size, const INT *zeros_start,
    INT tuple_size, R *real_data
    )
{
  INT m, mo, no[3], zeros_end[3];
  INT k0, k1, k2, h;
  int k0_is_inside, k1_is_inside, k2_is_inside, index_is_inside;
  
  PX(vadd_INT)(3, zeros_start, zeros_size, zeros_end);
  PX(vadd_INT)(3, zeros_size, n, no);

  if( PX(equal_INT)(3, zeros_start, zeros_end) )
    return;

  m  = tuple_size * PX(prod_INT)(3, n) - 1;
  mo = tuple_size * PX(prod_INT)(3, no) - 1;
  for(k0 = no[0] - 1 ; k0 >= 0; k0--){
    k0_is_inside = (zeros_start[0] <= k0) && (k0 < zeros_end[0]);
    for(k1 = no[1] - 1 ; k1 >= 0; k1--){
      k1_is_inside = (zeros_start[1] <= k1) && (k1 < zeros_end[1]);
      for(k2 = no[2] - 1 ; k2 >= 0; k2--){
        k2_is_inside = (zeros_start[2] <= k2) && (k2 < zeros_end[2]);
        index_is_inside = k0_is_inside || k1_is_inside || k2_is_inside;
        for(h = 0; h < tuple_size; h++)
          real_data[mo--] = (index_is_inside) ? 0 : real_data[m--];
      }
    }
  }
}


void PX(truncate_array_inside)(
    const INT *n, const INT *trunc_size, const INT *trunc_start,
    INT tuple_size, R *real_data
    )
{
  INT n_total, m, mo, k0, k1, k2, h, trunc_end[3];
  int k0_is_inside, k1_is_inside, k2_is_inside, index_is_inside;

  PX(vadd_INT)(3, trunc_start, trunc_size, trunc_end);

  if( PX(equal_INT)(3, trunc_start, trunc_end) )
    return;

  m = mo = 0;
  for(k0 = 0; k0 < n[0]; k0++){
    k0_is_inside = (trunc_start[0] <= k0) && (k0 < trunc_end[0]);
    for(k1 = 0; k1 < n[1]; k1++){
      k1_is_inside = (trunc_start[1] <= k1) && (k1 < trunc_end[1]);
      for(k2 = 0; k2 < n[2]; k2++){
        k2_is_inside = (trunc_start[2] <= k2) && (k2 < trunc_end[2]);
        index_is_inside = k0_is_inside || k1_is_inside || k2_is_inside;
        for(h = 0; h < tuple_size; h++){
          if( !index_is_inside )
            real_data[mo++] = real_data[m];
          m++;
        }
      }
    }
  }

  n_total = tuple_size * PX(prod_INT)(3, n);
  for(k0 = mo; k0 < n_total; k0++)
    real_data[k0] = 0;
}



int PX(pos_mod)(int dividend, int divisor)
{
  /* wrapper for C modulo function with positive result */

  /* assure -divisor <= dividend < divisor */
  dividend %= divisor;
  return (dividend < 0) ? dividend + ABS(divisor) : dividend;
}


int PX(get_mpi_cart_coord)(MPI_Comm comm_cart_1d, int *coord)
{
  int rank;
  MPI_Comm_rank(comm_cart_1d, &rank);
  return MPI_Cart_coords(comm_cart_1d, rank, 1, coord);
}

int PX(get_mpi_cart_dims)(MPI_Comm comm_cart, int maxdims, int *dims)
{
  int ret, *periods, *coords;
  
  periods = (int*) malloc(sizeof(int)*maxdims);
  coords  = (int*) malloc(sizeof(int)*maxdims);
  ret =  MPI_Cart_get(comm_cart, maxdims, dims, periods, coords);
  free(periods);
  free(coords);
  return ret;
}

void evaluate_user_block_size_3d(
    const INT *n, const INT *block, MPI_Comm comm,
    INT *block_intern
    )
{
  int num_procs;
  MPI_Comm_size(comm, &num_procs);
  
  for(int t=0; t<3; t++){
    if(block == PFFT_DEFAULT_BLOCKS)
      block_intern[t] = PX(global_block_size)(n[t], PFFT_DEFAULT_BLOCK, num_procs);
    else
      block_intern[t] = PX(global_block_size)(n[t], block[t], num_procs);
  }
}


void PX(evaluate_user_block_size_3d_truncated)(
    const INT *ni, const INT *no, const PX(permutation) *perm,
    const INT *block_col, const INT *block_row,
    MPI_Comm comm_col, MPI_Comm comm_row,
    INT *block_col_intern, INT *block_row_intern
    )
{
  int first_dim_col, first_dim_row;
  INT n_col[3], n_row[3];
  
  first_dim_col = PX(get_dim_to_comm_num)(perm, 0);
  first_dim_row = PX(get_dim_to_comm_num)(perm, 1);

  for(int t=0; t<3; t++){
    n_col[t] = (t == perm->n[first_dim_col]) ? ni[t] : no[t];
    n_row[t] = (t == perm->n[first_dim_row]) ? ni[t] : no[t];
  }
  
  evaluate_user_block_size_3d(n_col, block_col, comm_col, block_col_intern);
  evaluate_user_block_size_3d(n_row, block_row, comm_row, block_row_intern);
}


void PX(evaluate_user_gcells)(
    const INT *gcells_below, const INT *gcells_above,
    INT *gcells_below_intern, INT *gcells_above_intern
    )
{
  for(int t=0; t<3; t++){
    gcells_below_intern[t] = (gcells_below == PFFT_NO_GCELLS) ? 0 : gcells_below[t];
    gcells_above_intern[t] = (gcells_above == PFFT_NO_GCELLS) ? 0 : gcells_above[t];
  }
}


void PX(copy_permutation)(
    const PX(permutation) *perm,
    PX(permutation) *copy
    )
{
  for(int t=0; t<3; t++){
    copy->n[t]  = perm->n[t];
    copy->np[t] = perm->np[t];
  }
}

int PX(remap_num_comm)(
    int first_dims, unsigned remap_flag, const PX(permutation) *perm
    )
{
  int dim = PX(flag_active)(remap_flag, FFTW_MPI_TRANSPOSED_IN) ? first_dims : 0;
  return perm->np[dim];
}


int PX(get_local_dim)(
    const PX(permutation) *perm
    )
{
  int comm_num = 2;
  return PX(get_dim_to_comm_num)(perm, comm_num);
}


int PX(get_dim_to_comm_num)(
    const PX(permutation) *perm, int comm_num
    )
{
  return PX(permute_inverse)(perm->np, comm_num);
}


int PX(permute_inverse)(
    const int *perm, int search_item
    )
{
  int retval = -1;

  for(int p=0; p<3; p++)
    if( perm[p] == search_item )
      retval = p;

  return retval;
}

void PX(get_permutation_output)(
    const PX(profile) *prf, const PX(permutation) *perm_in,
    PX(permutation) *perm_out
    )
{
  PX(copy_permutation)(perm_in, perm_out);
  for(int p=0; p<3; p++){
    PX(permute_fft_1d)(prf->fft_flag[p], perm_out);
    PX(permute_transpose_2d)(prf->first_dims[p], prf->remap_flag[p], perm_out);
  }
}

int PX(flag_active)(
    unsigned flags, unsigned search_flag
    )
{
  return ( flags & search_flag );
}


int PX(flag_not_active)(
    unsigned flags, unsigned search_flag
    )
{
  return ( (~flags) & search_flag );
}


int PX(is_illegal_plan)(
    pfft_plan fftplan
    )
{
  int retval = 0;

  for(int p=0; p<3; p++){
    if(fftplan->local_fft[p]==NULL){
      retval = 1;
    }
  }

  return retval;
}


INT PX(prod_INT)(
    int d, const INT *vec
    )
{
  INT prod, t;

  for(prod=1, t=0; t<d; t++)
    prod *= vec[t];

  return prod;
}


int PX(equal_INT)(
    int d, const INT *vec1, const INT *vec2
    )
{
  int equal, t;
  for(equal=1, t=0; t<d; t++)
    equal *= (vec1[t] == vec2[t]);

  return equal;
}

void PX(vcopy_INT)(
    int d, const INT *vec1,
    INT *vec2
    )
{
  for(int t=0; t<d; t++)
    vec2[t] = vec1[t];
}


void PX(vadd_INT)(
    int d, const INT *vec1, const INT *vec2,
    INT *sum
    )
{
  for(int t=0; t<d; t++)
    sum[t] = vec1[t] + vec2[t];
}

void PX(vsub_INT)(
    int d, const INT *vec1, const INT *vec2,
    INT *sum
    )
{
  for(int t=0; t<d; t++)
    sum[t] = vec1[t] - vec2[t];
}

void PX(vfprintf)(
    MPI_Comm comm, FILE *restrict stream,
    const char *restrict format, va_list ap
    )
{
  int myrank;
  MPI_Comm_rank(comm, &myrank);
  if(!myrank)
    vfprintf(stream, format, ap);
}

void PX(fprintf)(
    MPI_Comm comm, FILE *restrict stream, const char *restrict format, ...
    )
{
  va_list ap;
  va_start(ap,format);
  PX(vfprintf)(comm, stream, format, ap);
  va_end(ap);
}

void PX(printf)(
    MPI_Comm comm, const char *restrict format, ...
    )
{
  va_list ap;
  va_start(ap,format);
  PX(vfprintf)(comm, stdout, format, ap);
  va_end(ap);
}


void PX(copy_profile)(
    const PX(profile) *prf,
    PX(profile) *copy
    )
{
  for(int t=0; t<3; t++){
    copy->fft_flag[t]  = prf->fft_flag[t];
    copy->first_dims[t] = prf->first_dims[t];
    copy->remap_flag[t] = prf->remap_flag[t];
  }
}
